; to be compiled with avra

.include "tn24adef.inc"

; bits of the port A
.equ	PCD_SCLK	=0	;PCD8544 serial clock
.equ	PCD_DC		=1	;PCD8544 data/command
.equ	PCD_SCE		=2	;PCD8544 chip enable
.equ	PCD_RES		=3	;PCD8544 reset
.equ	MK_CLOCK	=4	;MK85 serial clock (USI serial clock)
.equ	MK_DATA		=6	;MK85 serial data (USI data input)
.equ	PCD_SDIN	=7	;PCD8544 serial data, do not change!

; bits of the port B
.equ	MK_SYNC		=2	;MK85 address/data select signal (INT0)

.def	Temp2		=r0	;general purpose variable
.def	SavePort	=r1
.def	IntData		=r2	;used only in an interrupt routine
.def	SaveSreg	=r3
.def	Temp1		=r16	;general purpose variable
.def	UsiSrVal	=r17	;fixed constant
.def	UsiCrVal	=r18	;fixed constant
.def	Column		=r19
.def	Row		=r20
.def	Position	=r21
.def	Cursor		=r22

; The register Z is used only in the interrupt routines to access the
; lookup table.
; The register Y is used only in the interrupt routines to access the
; RAM buffer.

; The data buffer occupies the RAM address range 0x60-0xC1.
; Data are stored on lower 5 bits, inverted and in reversed order.
; The location 0xC0 stores the cursor.
; The location 0xC1 is a sink for addresses out of the range 0x60-0xC0.
.equ	buffer		=0x60


; expects two immediate coordinates: Y, X
.macro		coords
		cbi	PORTA,PCD_DC	;command
		ldi	Temp1,0x40 | @0	;Y-address
		rcall	write_1
		ldi	Temp1,0x80 | @1	;X-address
		rcall	write_1
.endm


; send an immediate value to the LCD
.macro		write
		ldi	Temp1,@0
		rcall	write_1
.endm


; clear a number of bits specified by an immediate value
.macro		clear
		ldi	Temp1,@0
		rcall	clear_1
.endm


		.org	0x0000

vector01:	rjmp	reset		;Reset Handler
; The IRQ0 Handler reads the received LCD address byte.
; It needs to be as fast as possible, hence no RJMP.
; Also the order of instructions is intentional, as we have more time to read
; USIDR (up to the next falling edge of the MK_SHIFT signal) than to write
; USISR (up to the next rising edge of the MK_SHIFT signal).
vector02:	rjmp	ext_int0	;IRQ0 Handler
vector03:	reti			;PCINT0 Handler
vector04:	reti			;PCINT1 Handler
vector05:	reti			;Watchdog Interrupt Handler
vector06:	reti			;Timer1 Capture Handler
vector07:	reti			;Timer1 Compare A Handler
vector08:	reti			;Timer1 Compare B Handler
vector09:	reti			;Timer1 Overflow Handler
vector10:	reti			;Timer0 Compare A Handler
vector11:	reti			;Timer0 Compare B Handler
vector12:	reti			;Timer0 Overflow Handler
vector13:	reti			;Analog Comparator Handler
vector14:	reti			;ADC Conversion Handler
vector15:	reti			;EEPROM Ready Handler
vector16:	reti			;USI Start Handler
vector17:	in	ZL,USIBR	;USI Overflow Handler
		lpm	YL,Z
		reti

; called at the rising edge of the MK_SYNC signal
ext_int0:	in	IntData,USIDR
		in	SaveSreg,SREG
		sbic	USISR,USICNT2
		lsr	IntData		;fix a possible glitch of the SHIFT signal
		out	USISR,UsiSrVal	;clear the 4-bit counter and its overflow flag
		st	Y,IntData
		ldi	YL,0xC1
		out	SREG,SaveSreg
		reti


		.org	0x0080

; An lookup table of inverted and reversed address bytes, then converted to
; a RAM address (values out of the range 0x80-0xE0 are replaced with 0xC1).
; It needs to be aligned to a 256-byte boundary.
rev:		.db	0xC1, 0xC1, 0x9F, 0xC1, 0xBF, 0xC1, 0x7F, 0xC1
		.db	0xC1, 0xC1, 0x8F, 0xC1, 0xAF, 0xC1, 0x6F, 0xC1
		.db	0xC1, 0xC1, 0x97, 0xC1, 0xB7, 0xC1, 0x77, 0xC1
		.db	0xC1, 0xC1, 0x87, 0xC1, 0xA7, 0xC1, 0x67, 0xC1
		.db	0xC1, 0xC1, 0x9B, 0xC1, 0xBB, 0xC1, 0x7B, 0xC1
		.db	0xC1, 0xC1, 0x8B, 0xC1, 0xAB, 0xC1, 0x6B, 0xC1
		.db	0xC1, 0xC1, 0x93, 0xC1, 0xB3, 0xC1, 0x73, 0xC1
		.db	0xC1, 0xC1, 0x83, 0xC1, 0xA3, 0xC1, 0x63, 0xC1
		.db	0xC1, 0xC1, 0x9D, 0xC1, 0xBD, 0xC1, 0x7D, 0xC1
		.db	0xC1, 0xC1, 0x8D, 0xC1, 0xAD, 0xC1, 0x6D, 0xC1
		.db	0xC1, 0xC1, 0x95, 0xC1, 0xB5, 0xC1, 0x75, 0xC1
		.db	0xC1, 0xC1, 0x85, 0xC1, 0xA5, 0xC1, 0x65, 0xC1
		.db	0xC1, 0xC1, 0x99, 0xC1, 0xB9, 0xC1, 0x79, 0xC1
		.db	0xC1, 0xC1, 0x89, 0xC1, 0xA9, 0xC1, 0x69, 0xC1
		.db	0xC1, 0xC1, 0x91, 0xC1, 0xB1, 0xC1, 0x71, 0xC1
		.db	0xC1, 0xC1, 0x81, 0xC1, 0xA1, 0xC1, 0x61, 0xC1
		.db	0xC1, 0xC1, 0x9E, 0xC1, 0xBE, 0xC1, 0x7E, 0xC1
		.db	0xC1, 0xC1, 0x8E, 0xC1, 0xAE, 0xC1, 0x6E, 0xC1
		.db	0xC1, 0xC1, 0x96, 0xC1, 0xB6, 0xC1, 0x76, 0xC1
		.db	0xC1, 0xC1, 0x86, 0xC1, 0xA6, 0xC1, 0x66, 0xC1
		.db	0xC1, 0xC1, 0x9A, 0xC1, 0xBA, 0xC1, 0x7A, 0xC1
		.db	0xC1, 0xC1, 0x8A, 0xC1, 0xAA, 0xC1, 0x6A, 0xC1
		.db	0xC1, 0xC1, 0x92, 0xC1, 0xB2, 0xC1, 0x72, 0xC1
		.db	0xC1, 0xC1, 0x82, 0xC1, 0xA2, 0xC1, 0x62, 0xC1
		.db	0xC1, 0xC1, 0x9C, 0xC1, 0xBC, 0xC1, 0x7C, 0xC1
		.db	0xC1, 0xC1, 0x8C, 0xC1, 0xAC, 0xC1, 0x6C, 0xC1
		.db	0xC1, 0xC1, 0x94, 0xC1, 0xB4, 0xC1, 0x74, 0xC1
		.db	0xC1, 0xC1, 0x84, 0xC1, 0xA4, 0xC1, 0x64, 0xC1
		.db	0xC1, 0xC1, 0x98, 0xC1, 0xB8, 0xC1, 0x78, 0xC1
		.db	0xC1, 0xC1, 0x88, 0xC1, 0xA8, 0xC1, 0x68, 0xC1
		.db	0xC1, 0xC1, 0x90, 0xC1, 0xB0, 0xC1, 0x70, 0xC1
		.db	0xC0, 0xC1, 0x80, 0xC1, 0xA0, 0xC1, 0x60, 0xC1


; Main program start
reset:
; the stack pointer doesn't need to be initialised
		ldi	XH,high(buffer)	;points to the RAM buffer
		ldi	YH,high(buffer)	;points to the RAM buffer
		ldi	ZH,high(rev<<1)	;points to the lookup table in the ROM

; port setup
		ldi	Temp1,(1<<PCD_SDIN) | (1<<PCD_SCLK) | (1<<PCD_DC) | (1<<PCD_RES) | (1<<PCD_SCE)
		out	DDRA,Temp1
		ldi	Temp1,(1<<PCD_SCE) | (1<<PCD_RES)
		out	PORTA,Temp1
		ldi	Temp1,0
		out	DDRB,Temp1
		ldi	Temp1,~(1<<MK_SYNC)	;pull-ups on unused pins
		out	PORTB,Temp1

; USI setup
; Start Condition Interrupt Disable
; Counter Overflow Interrupt Enable
; port pins operate as normal
; external clock, negative edge
		ldi	UsiCrVal,(0<<USISIE) | (1<<USIOIE) | (0<<USIWM1) | (0<<USIWM0) | (1<<USICS1) | (1<<USICS0) | (0<<USICLK) | (0<<USITC)
		out	USICR,UsiCrVal
; clear the 4-bit counter and its overflow flag
		ldi	UsiSrVal,(1<<USIOIF) | (0<<USICNT3) | (0<<USICNT2) | (0<<USICNT1) | (0<<USICNT0)
		out	USISR,UsiSrVal

; external INT0 setup
		ldi	Temp1,(1<<ISC01) | (1<<ISC00)	;rising edge
		out	MCUCR,Temp1
		ldi	Temp1,(1<<INT0)
		out	GIMSK,Temp1
		sei

; timer1 setup
		ldi	Temp1,0		;normal mode, WMG1[3:0] = 0000
		out	TCCR1A,Temp1
		ldi	Temp1,(0<<CS12) | (1<<CS11) | (1<<CS10)	;prescaler = 64
		out	TCCR1B,Temp1

; PCD8544 reset pulse 1ms
		clr	Temp1
		out	TCNT1H,Temp1
		out	TCNT1L,Temp1
delay_1:	in	Temp1,TCNT1L
		in	Temp2,TCNT1H
		sbrs	Temp1,7
		rjmp	delay_1
		cbi	PORTA,PCD_RES
delay_2:	in	Temp1,TCNT1L
		in	Temp2,TCNT1H
		sbrc	Temp1,7
		rjmp	delay_2
		sbi	PORTA,PCD_RES

		write	0x21		;function set, extended
		write	0x06		;temperature  coefficient = 2
		write	0x13		;bias = 4
		write	0xC2		;contrast = 0x42

		write	0x20		;function set, normal
		write	0x0C		;normal display mode

; clear the display RAM
		coords	0,0
		ldi	XL,18		;number_of_rows*3 = 6*3 = 18
cl1:		clear	28*8		;number_of_columns/3 = 84/3 = 28
		dec	XL
		brne	cl1

; show the buffer contents
loop:		coords	0,0
		lds	Temp1,buffer+0x00
		andi	Temp1,0x10
		rcall	message_ext

		coords	0,20
		lds	Temp1,buffer+0x00
		andi	Temp1,0x08
		rcall	message_s

		coords	1,20
		lds	Temp1,buffer+0x00
		andi	Temp1,0x04
		rcall	message_f

		coords	0,28
		lds	Temp1,buffer+0x08
		andi	Temp1,0x10
		rcall	message_run

		coords	1,28
		lds	Temp1,buffer+0x08
		andi	Temp1,0x08
		rcall	message_wrt

		coords	0,48
		lds	Temp1,buffer+0x08
		andi	Temp1,0x01
		rcall	message_deg

		coords	1,48
		lds	Temp1,buffer+0x18
		andi	Temp1,0x10
		rcall	message_rad

		coords	0,67
		lds	Temp1,buffer+0x20
		andi	Temp1,0x10
		rcall	message_gra

		coords	1,70
		lds	Temp1,buffer+0x28
		andi	Temp1,0x10
		rcall	message_tr

		coords	2,13
		rcall	sevenseg

		coords	2,53
		lds	Temp1,buffer+0x58
		andi	Temp1,0x02
		rcall	message_stop

		coords	4,0
		rcall	text

		rjmp	loop


; send a byte Temp1 to the LCD
; PCD_SDIN is required to be on the bit 7 of the port A
write_1:	cbi	PORTA,PCD_SCE
		in	SavePort,PORTA
		lsl	SavePort
		sec
		rol	Temp1
write_2:	ror	SavePort
		out	PORTA,SavePort
		lsl	SavePort
		lsl	Temp1
		sbi	PORTA,PCD_SCLK
		brne	write_2
		cbi	PORTA,PCD_SCLK
		cbi	PORTA,PCD_SDIN
		sbi	PORTA,PCD_SCE
		ret


; display the main field of 12 characters 5x7
text:		cbi	PORTA,PCD_SCE
		sbi	PORTA,PCD_DC	;data
		in	SavePort,PORTA
		lsl	SavePort
		ldi	XL,low(buffer)
		ldi	Position,0
text_pos_loop:	ldi	Cursor,0	;no cursor
		in	Temp1,TCNT1L
		in	Temp1,TCNT1H
		tst	Temp1
		brpl	text_skip_crsr
		ldi	Column,0	;variable reused for bit reversal
		lds	Temp1,buffer+0x60
		sbrs	Temp1,1
		sbr	Column,0x08
		sbrs	Temp1,2
		sbr	Column,0x04
		sbrs	Temp1,3
		sbr	Column,0x02
		sbrs	Temp1,4
		sbr	Column,0x01	;Column = reversed and inverted Temp1
		cp	Column,Position
		brne	text_skip_crsr
		ldi	Cursor,8	;underscore
		sbrc	Temp1,0
		ldi	Cursor,1	;block
text_skip_crsr:	ldi	Column,0x40
text_col_loop:	subi	XL,-8
		ldi	Temp1,0xFF	;blank
		ldi	Row,8
text_row_loop:	and	Temp1,Column
		cpi	Temp1,1		;Carry set if Temp1 = 0
		ror	SavePort
		out	PORTA,SavePort
		lsl	SavePort
		dec	XL
		cpi	Column,0x11
		brcc	text_skip_data	;spaces between characters
		tst	Cursor
		breq	text_get_data
		ldi	Temp1,0xFF	;blank
		cp	Row,Cursor
		brcs	text_skip_data
		ldi	Temp1,0x00	;pixels set
		rjmp	text_skip_data
text_get_data:	ld	Temp1,X
text_skip_data:	sbi	PORTA,PCD_SCLK
		dec	Row
		brne	text_row_loop	;next row
		lsr	Column
		brne	text_col_loop	;next column
		subi	XL,-8
		inc	Position
		cpi	Position,12
		brcs	text_pos_loop	;next character
		cbi	PORTA,PCD_SCLK
		cbi	PORTA,PCD_SDIN
		sbi	PORTA,PCD_SCE
		ret


; display the four 7-segment digits
sevenseg:	sbi	PORTA,PCD_DC	;data
; 1st digit
		ldi	Temp1,0
		lds	XL,buffer+0x30
		sbrs	XL,4
		sbr	Temp1,0x06
		sbrs	XL,2
		sbr	Temp1,0x30
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,3
		sbr	Temp1,0x08
		sbrs	XL,1
		sbr	Temp1,0x40
		lds	XL,buffer+0x38
		sbrs	XL,3
		sbr	Temp1,0x01
		mov	Temp2,Temp1
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,4
		sbr	Temp1,0x06
		lds	XL,buffer+0x30
		sbrs	XL,0
		sbr	Temp1,0x30
		rcall	write_1
		write	0
		write	0
; 2nd digit
		ldi	Temp1,0
		lds	XL,buffer+0x38
		sbrs	XL,2
		sbr	Temp1,0x06
		sbrs	XL,0
		sbr	Temp1,0x30
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,1
		sbr	Temp1,0x08
		lds	XL,buffer+0x40
		sbrs	XL,4
		sbr	Temp1,0x40
		sbrs	XL,1
		sbr	Temp1,0x01
		mov	Temp2,Temp1
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,2
		sbr	Temp1,0x06
		sbrs	XL,3
		sbr	Temp1,0x30
		rcall	write_1
		write	0
		write	0
; 3rd digit
		ldi	Temp1,0
		sbrs	XL,0
		sbr	Temp1,0x06
		lds	XL,buffer+0x48
		sbrs	XL,3
		sbr	Temp1,0x30
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,4
		sbr	Temp1,0x08
		sbrs	XL,2
		sbr	Temp1,0x40
		lds	XL,buffer+0x50
		sbrs	XL,4
		sbr	Temp1,0x01
		mov	Temp2,Temp1
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		ldi	Temp1,0
		lds	XL,buffer+0x48
		sbrs	XL,0
		sbr	Temp1,0x06
		sbrs	XL,1
		sbr	Temp1,0x30
		rcall	write_1
		write	0
		write	0
; 4th digit
		ldi	Temp1,0
		lds	XL,buffer+0x50
		sbrs	XL,3
		sbr	Temp1,0x06
		sbrs	XL,1
		sbr	Temp1,0x30
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,2
		sbr	Temp1,0x08
		sbrs	XL,0
		sbr	Temp1,0x40
		lds	XL,buffer+0x58
		sbrs	XL,2
		sbr	Temp1,0x01
		mov	Temp2,Temp1
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		mov	Temp1,Temp2
		rcall	write_1
		ldi	Temp1,0
		sbrs	XL,3
		sbr	Temp1,0x06
		sbrs	XL,4
		sbr	Temp1,0x30
		rcall	write_1
		ret


message_ext:	ldi	Temp1,17*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_E
		rcall	write_X
		rcall	write_T
		ret

message_f:	ldi	Temp1,5*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_F
		ret

message_s:	ldi	Temp1,5*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_S
		ret

message_run:	ldi	Temp1,17*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_R
		rcall	write_U
		rcall	write_N
		ret

message_wrt:	ldi	Temp1,17*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_W
		rcall	write_R
		rcall	write_T
		ret


; clear a number of bits specified in Temp1
clear_1:	cbi	PORTA,PCD_SCE
		cbi	PORTA,PCD_SDIN
		sbi	PORTA,PCD_DC	;data
clear_2:	sbi	PORTA,PCD_SCLK
		dec	Temp1
		cbi	PORTA,PCD_SCLK
		brne	clear_2
		sbi	PORTA,PCD_SCE
		ret


message_deg:	ldi	Temp1,17*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_D
		rcall	write_E
		rcall	write_G
		ret

message_rad:	ldi	Temp1,17*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_R
		rcall	write_A
		rcall	write_D
		ret

message_gra:	ldi	Temp1,17*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_G
		rcall	write_R
		rcall	write_A
		ret

message_tr:	ldi	Temp1,11*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_T
		rcall	write_R
		ret

message_stop:	ldi	Temp1,23*8
		brne	clear_1
		sbi	PORTA,PCD_DC	;data
		rcall	write_S
		rcall	write_T
		rcall	write_O
		rcall	write_P
		ret


; letter 'A'
write_A:	write	0x7E
		write	0x11
		write	0x11
		write	0x11
		write	0x7E
		write	0x00
		ret

; letter 'D'
write_D:	write	0x7F
		write	0x41
		write	0x41
		write	0x22
		write	0x1C
		write	0x00
		ret

; letter 'E'
write_E:	write	0x7F
		write	0x49
		write	0x49
		write	0x49
		write	0x41
		write	0x00
		ret

; letter 'F'
write_F:	write	0x7F
		write	0x09
		write	0x09
		write	0x09
		write	0x01
		write	0x00
		ret

; letter 'G'
write_G:	write	0x3E
		write	0x41
		write	0x49
		write	0x49
		write	0x7A
		write	0x00
		ret

; letter 'N'
write_N:	write	0x7F
		write	0x04
		write	0x08
		write	0x10
		write	0x7F
		write	0x00
		ret

; letter 'O'
write_O:	write	0x3E
		write	0x41
		write	0x41
		write	0x41
		write	0x3E
		write	0x00
		ret

; letter 'P'
write_P:	write	0x7F
		write	0x09
		write	0x09
		write	0x09
		write	0x06
		write	0x00
		ret

; letter 'R'
write_R:	write	0x7F
		write	0x09
		write	0x19
		write	0x29
		write	0x46
		write	0x00
		ret

; letter 'S'
write_S:	write	0x46
		write	0x49
		write	0x49
		write	0x49
		write	0x31
		write	0x00
		ret

; letter 'T'
write_T:	write	0x01
		write	0x01
		write	0x7F
		write	0x01
		write	0x01
		write	0x00
		ret

; letter 'U'
write_U:	write	0x3F
		write	0x40
		write	0x40
		write	0x40
		write	0x3F
		write	0x00
		ret

; letter 'W'
write_W:	write	0x3F
		write	0x40
		write	0x38
		write	0x40
		write	0x3F
		write	0x00
		ret

; letter 'X'
write_X:	write	0x63
		write	0x14
		write	0x08
		write	0x14
		write	0x63
		write	0x00
		ret
